{
open Parser
open Parsetree

type tag_type = Open | Closing

let tag_list =
  [("B", TB);
   ("EM", TEM);
   ("I", TI);
   ("PL", TPL);
   ("S", TS);
   ("TT", TTT);
   ("U", TU);
   ("0", TSize 0);
   ("1", TSize 1);
   ("2", TSize 2);
   ("3", TSize 3);
   ("4", TSize 4);
   ("5", TSize 5);
   ("6", TSize 6);
   ("7", TSize 7);
   ("8", TSize 8);
   ("9", TSize 9);
   ("r", TColor R);
   ("g", TColor G);
   ("b", TColor B);
   ("c", TColor C);
   ("m", TColor M);
   ("y", TColor Y);
   ("k", TColor K);
   ("w", TColor W)]

let tag_hash = Hashtbl.create 29

let _ =
  let add_tag tag_str tag_cons =
    Hashtbl.add tag_hash tag_str tag_cons in
  List.iter2 add_tag 
    (List.map fst tag_list)
    (List.map snd tag_list)

let tag_token tag tag_type =
  try
    let tag_cons = Hashtbl.find tag_hash tag in
    match tag_type with
      Open -> OTAG (tag_cons, tag)
    | Closing -> CTAG (tag_cons, tag)
  with
    Not_found ->
      failwith ("<" ^ tag ^ "> is not a valide tag")

let string_of_otag s =
  String.sub s 1 (String.length s - 2)

let string_of_ctag s =
  String.sub s 2 (String.length s - 3)

} 

let blank = ['\032' '\013' '\010' '\009']
let textchar = ['\033'-'\059' '\061' '\063'-'\126']

rule lexer = parse
    blank
    { BLANK (Lexing.lexeme_char lexbuf 0) }
| "</" textchar* '>'
    { tag_token (string_of_ctag (Lexing.lexeme lexbuf)) Closing }
| '<' textchar* '>'
    { tag_token (string_of_otag (Lexing.lexeme lexbuf)) Open }
| ['<' '>'] 
    { failwith ("Invalid use of " ^ (Lexing.lexeme lexbuf)) }
| textchar+
    { TEXT (Lexing.lexeme lexbuf) }
| eof { EOF }
| _
    { failwith (Lexing.lexeme lexbuf ^ " is not a valid character") }
