open Parsetree
open State

(* checked *)

let new_state state = function
    | TB ->
	{state with b = true}
    | TEM ->
	if not state.s then {state with em = not state.em} else state
    | TI ->
	{state with i = true}
    | TPL ->
	{state with u = 0; s = false; em = false; i = false; b = false;
	 tt = false}
    | TS ->
	{state with s = true; em = false}
    | TTT ->
	{state with tt = true}
    | TU ->
	if state.u < 3 then {state with u = state.u + 1} else state
    | TSize size ->
	{state with size = size}
    | TColor c ->
	{state with color = c}
    | Notag ->
	state

(* checked *)

let rec simp_red state = function
    Tag (tag, s, d) :: t ->
      let state2 = new_state state tag in
      let sd = simp_red state2 d in
      (if state = state2 then
	Tag (Notag, "", sd)
      else
	Tag (tag, s, sd)) :: simp_red state t
  | Space c :: t -> Space c :: simp_red state t
  | Text s :: t -> Text s :: simp_red state t
  | Nil :: t -> simp_red state t
  | [] -> []

(* checked *)

let space_state state =
  {state with s = false; em = false; i = false; b = false;
   color = if state.u = 0 then W else state.color}

(* checked *)

let space_state_equiv e1 e2 =
  space_state e1 = space_state e2 && e1.tt = false

(* checked *)

let first_char_info state tree =
  let rec first_char_aux state = function
      Tag (tag, s, d) :: t ->
	let found_char, is_fc_spc, state2, wo_spc = 
	  first_char_aux (new_state state tag) d in
	if not found_char then 
	  let found_char, is_fc_spc, state2, wo_spc = 
	    first_char_aux state t in
	  (found_char, is_fc_spc, state2, Tag (tag, s, d) :: wo_spc)
	else
	(true, is_fc_spc, state2, Tag (tag, s, wo_spc) :: t)
    | Space _ :: t -> (true, true, state, t)
    | Text _ :: t -> (true, false, state, [])
    | Nil :: t -> first_char_aux state t
    | [] -> (false, false, state, []) in
  let _, is_fc_spc, state2, wo_spc = first_char_aux state tree in
  (is_fc_spc, state2, wo_spc)

(* checked *)

let rec simp_space state = function
    Space c :: [] ->
      (Space c :: [], true, state)
  | Text t :: [] ->
      (Text t :: [], false, state)
  | Space c :: t ->
      let st, is_lc_spc, lc_state = simp_space state t in
      let is_fc_spc, fc_state, _ = first_char_info state st in
      if is_fc_spc && space_state_equiv state fc_state then
	(st, is_lc_spc, lc_state)
      else
	(Space c :: st, is_lc_spc, lc_state)
  | Tag (tag, s, d) :: t ->
      let sd, d_is_lc_spc, d_state = simp_space state d
      and st, t_is_lc_spc, t_state = simp_space state t in
      let st, t_is_lc_spc, t_state =
	if st = [] then
	  (st, d_is_lc_spc, d_state)
	else
	  (st, t_is_lc_spc, t_state)
      and is_fc_spc, state2, wo_spc = first_char_info state st in
      if d_is_lc_spc && is_fc_spc && space_state_equiv d_state state2
      then
	(Tag (tag, s, sd) :: wo_spc, t_is_lc_spc, t_state)
      else
	(Tag (tag, s, sd) :: st, t_is_lc_spc, t_state)
  | Text s :: t ->
      let st, is_lc_spc, lc_state = simp_space state t in
      (Text s :: st, is_lc_spc, lc_state)
  | Nil :: t -> simp_space state t
  | [] -> ([], false, def_state)

(* checked *)

let rec is_em_tag = function
    Tag (Notag, _, d) :: t -> is_em_tag d
  | Tag (TEM, _, d) :: t -> (true, d)
  | Nil :: t -> is_em_tag t
  | _ -> (false, [])

(* checked *)

let rec do_simp_em = function
  | Tag (TEM, s, d) :: t ->
      let sd, st = do_simp_em d, do_simp_em t in
      let em = is_em_tag sd in
      if fst em then
	Tag (Notag, "", snd em) :: st
      else
	Tag (TEM, s, sd) :: st
  | Tag (tag, s, d) :: t ->
      Tag (tag, s, do_simp_em d) :: do_simp_em t
  | h :: t -> h :: do_simp_em t
  | [] -> []

let do_simp_space tree =
  match simp_space def_state tree with
    stree, _, _ -> stree

let do_simp_red tree =
  simp_red def_state tree

let simp tree =
  do_simp_em (do_simp_space (do_simp_red tree))
